/*
  This code was written by Raymond Chi-Wing Wong.
  Please read the following.

    Raymond Chi-Wing Wong, Yufei Tao, Ada Wai-Chee Fu and Xiaokui Xiao. 
    "On Efficient Spatial Matching", 
    the 33rd International Conference on Very Large Data Bases (VLDB'07), 
    Vienna, Austria on 23-28 Sept, 2007 

*/

#include "assignment.h"


triple *triple_new()
{
	triple *returnValue;

	returnValue = (triple *) malloc(sizeof(triple));
	memset(returnValue, 0, sizeof(triple));

	return returnValue;
}

triple *triple_newValue(int oid, int pid, int weight)
{
	triple *returnValue;

	returnValue = triple_new();

	returnValue->oid = oid;
	returnValue->pid = pid;
	returnValue->weight = weight;

	return returnValue;
}

void triple_free(triple **aTriple)
{
	free(*aTriple);
}

assignment *assignment_new()
{
	assignment *returnValue;

	returnValue = (assignment *) malloc(sizeof(assignment));
	memset(returnValue, 0, sizeof(assignment));

	return returnValue;
}

void assignment_free(assignment **aAssignment)
{
	int noOfMatch;
	triple **list;
	int i;

	noOfMatch = (*aAssignment)->noOfMatch;
	list = (*aAssignment)->list;
	if (noOfMatch > 0)
	{
		for (i = 0; i < noOfMatch; i++)
		{
			triple_free(&(list[i]));
		}
		free(list);
	}
	free(*aAssignment);
}

void assignment_insert(assignment *aAssignment, triple **aTriple)
{
	int noOfMatch;
	triple **list;

	noOfMatch = aAssignment->noOfMatch;
	list = aAssignment->list;
	if (noOfMatch == 0)
	{
		list = (triple **) malloc(sizeof(triple *));
	}
	else
	{
		list = (triple **) realloc(list, sizeof(triple *)*(noOfMatch+1));
	}
	list[noOfMatch] = (*aTriple);

	aAssignment->noOfMatch++;
	aAssignment->list = list;
}


void assignment_print(char *filename, assignment *aAssignment)
{
	FILE *filePtr;

	int noOfMatch;
	triple **list;
	triple *aTriple;

	int oid, pid, weight;

	int i;

	filePtr = (FILE *) fopen(filename, "w");
	if (filePtr == NULL)
	{
		printf("File %s cannot be written!\n", filename);
		exit(0);
	}

	noOfMatch = aAssignment->noOfMatch;
	list = aAssignment->list;

	for (i = 0; i < noOfMatch; i++)
	{
		aTriple = list[i];

		oid = aTriple->oid;
		pid = aTriple->pid;
		weight = aTriple->weight;

		fprintf(filePtr, "%d. (%d, %d, %d)\n", i, oid, pid, weight);
	}
	

	fclose(filePtr);
}

/// the content pointing from aAssignment[index] has not been removed
void assignment_removeIndex(assignment *aAssignment, int index)
{
	int noOfMatch;
	triple **list;

	triple **newList;

	noOfMatch = aAssignment->noOfMatch;
	list = aAssignment->list;

	if ((0 <= index) && (index < noOfMatch))
	{
		if (noOfMatch == 1)
		{
			newList = NULL;
		}
		else
		{
			newList = (triple **) malloc(sizeof(triple *)*(noOfMatch-1));
			if (index != 0)
			{
				memcpy(newList, list, sizeof(triple *)*index);
			}

			if (index != noOfMatch-1)
			{
				memcpy(&(newList[index]), &(list[index+1]), sizeof(triple *)*(noOfMatch-index-1));
			}
		}

		aAssignment->noOfMatch--;
		aAssignment->list = newList;

		free(list);
	}
}

assignment *assignment_union(assignment *aAssignment1, assignment *aAssignment2)
{
	assignment *returnValue;

	int noOfMatch1, noOfMatch2, noOfMatchTotal;
	triple **list1;
	triple **list2;
	triple **listTotal;

	int i;

	returnValue = assignment_new();

	noOfMatch1 = aAssignment1->noOfMatch;
	list1 = aAssignment1->list;

	noOfMatch2 = aAssignment2->noOfMatch;
	list2 = aAssignment2->list;
	
	noOfMatchTotal = noOfMatch1 + noOfMatch2;

	if (noOfMatchTotal > 0)
	{
		listTotal = (triple **) malloc(sizeof(triple *)*noOfMatchTotal);

		for (i = 0; i < noOfMatch1; i++)
		{
			listTotal[i] = list1[i];
		}

		for (i = 0; i < noOfMatch2; i++)
		{
			listTotal[i+noOfMatch1] = list2[i];
		}
	}
	else
	{
		listTotal = NULL;
	}

	returnValue->noOfMatch = noOfMatchTotal;
	returnValue->list = listTotal;

	return returnValue;
}

assignmentSet *assignmentSet_new()
{
	assignmentSet *returnValue;

	returnValue = (assignmentSet *) malloc(sizeof(assignmentSet));
	memset(returnValue, 0, sizeof(assignmentSet));

	return returnValue;
}

void assignmentSet_free(assignmentSet **aAssignmentSet)
{
	int noOfAssign;
	assignment **list;
	int i;

	noOfAssign = (*aAssignmentSet)->noOfAssign;
	if (noOfAssign > 0)
	{
		list = (*aAssignmentSet)->list;
		for (i = 0; i < noOfAssign; i++)
		{
			assignment_free(&(list[i]));
		}
	}
	free((*aAssignmentSet));
}

void assignmentSet_insert(assignmentSet *aAssignmentSet, assignment **aAssignment)
{
	int noOfAssign;
	assignment **list;
	assignment **newList;

	noOfAssign = aAssignmentSet->noOfAssign;
	list = aAssignmentSet->list;

	if (noOfAssign == 0)
	{
		newList = (assignment **) malloc(sizeof(assignment *));
	}
	else
	{
		newList = (assignment **) realloc(list, sizeof(assignment *)*(noOfAssign+1));
	}
	newList[noOfAssign] = (*aAssignment);

	aAssignmentSet->noOfAssign++;
	aAssignmentSet->list = newList;
}


void assignmentSet_transformFromFinalAssignment(assignment *aAssignment, int noOfPointO, int noOfPointP, assignmentSet **assignmentSetForO, assignmentSet **assignmentSetForP)
{

	int noOfMatch;
	triple **list;
	triple *aTriple;

	assignment *pointAssignmentP;
	assignment *pointAssignmentO;

	assignmentSet *returnValueP;
	assignmentSet *returnValueO;

	int i;
	int pid, oid;

	//
	noOfMatch = aAssignment->noOfMatch;
	list = aAssignment->list;

	//
	returnValueO = assignmentSet_new();
	returnValueP = assignmentSet_new();

	returnValueO->noOfAssign = noOfPointO;
	returnValueO->list = (assignment **) malloc(sizeof(assignment *)*noOfPointO);
	for (i = 0; i < noOfPointO; i++)
	{
		returnValueO->list[i] = assignment_new();
	}

	returnValueP->noOfAssign = noOfPointP;
	returnValueP->list = (assignment **) malloc(sizeof(assignment *)*noOfPointP);
	for (i = 0; i < noOfPointP; i++)
	{
		returnValueP->list[i] = assignment_new();
	}
	
	for (i = 0; i < noOfMatch; i++)
	{	
		aTriple = list[i];

		oid = aTriple->oid;
		pid = aTriple->pid;

		pointAssignmentO = returnValueO->list[oid];
		pointAssignmentP = returnValueP->list[pid - noOfPointO];

		assignment_insert(pointAssignmentO, &aTriple);
		assignment_insert(pointAssignmentP, &aTriple);
	}

	(*assignmentSetForO) = returnValueO;
	(*assignmentSetForP) = returnValueP;
}



