

#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <fstream>
#include <iostream>
#include <algorithm>
#include "rand.h"

#include "rtree.h"

#include "MersenneTwister.h"





#define min(a, b) (((a) < (b))? (a) : (b)  )
#define max(a, b) (((a) > (b))? (a) : (b)  )

/*****************************************************************
1d dataset.
this function ensures that no two generated numbers are identical.

fname: output filename
n: cardinality
u: maximum value of the dimension

coded by yufei tao, may 2005.
*****************************************************************/
void one_dim(char *_fname, int _n, float _u)
{
	FILE *fp=fopen(_fname, "w");
	if (!fp)
	{
		printf("couldn't create file\n"); exit(1);
	}

	float *data = new float[_n];

	for (int i=0; i<_n; i++)
	{
		if ((i + 1) % 1000 == 0)
			printf("finished %d \n", i + 1);
		bool again = true;
		float x;
		while (again)
		{
			x = new_uniform(0, _u);

			again = false;
			//check if x has been generated before 
			for (int j = 0; j < i; j++)
				if (fabs(x - data[j]) < FLOATZERO)
					again = true;
		}

		fprintf(fp, "%d %f\n", i+1, x);
	}

	delete []data;
	fclose(fp);
}

/*****************************************************************
Multi-dimensional point dataset

fname: output filename
n: cardinality
u: universe
d: the dimensionality
dist: distribution (1 = uniform, 2 = zipf)

coded by yufei tao, may 2005.
*****************************************************************/
void spatial_ds(const char *_fname, int _n, int _d, float _u, int _dist)
{
	FILE *fp=fopen(_fname, "w");
	if (!fp)
	{
		printf("couldn't create file\n"); exit(1);
	}
	int dim=_d;
	for (int i=0; i<_n; i++)
	{
		fprintf(fp, "%d", i+1);
		for (int j=0; j<dim; j++)
		{
			float x0;
			if (_dist == 1)
				x0 = new_uniform(0, _u);

			if (_dist == 2)
			{
				x0=-1;
				while (x0<0 || x0>_u)
					x0=zipf(0, _u, 1);
			}

			fprintf(fp, " %.2f", x0);
		}
		fprintf(fp, "\n");
	}
	fclose(fp);
}




/*****************************************************************
Multi-dimensional correlated data
para
fname: output filename
n: cardinality (also decides the window size)
d: the dimensionality
u: universe
coded by Xiaokui Xiao, 20061128
*****************************************************************/

void correlated( const char *pFileName, int nCardin, int nDimen, float fAxisLen )
{
	MTRand temp_rand;
	temp_rand.seed();

	FILE* fp = fopen( pFileName, "w" );
	if (fp == NULL)
	{
		printf("correlated() error: couldn't create file\n");
		return;
	}

	float* x = new float[nDimen];
	for (int i = 0; i < nCardin; ++i)
	{
		fprintf(fp, "%d", i+1);
		
		bool shall_gen = true;
		float tmp;

		while (shall_gen == true)
		{			
			shall_gen = false;
			tmp = temp_rand.rand( fAxisLen );
			for (int j = 0; j < nDimen; ++j)
			{
				x[j] = tmp + temp_rand.randNorm( 0, 0.05 * 0.05 ) * fAxisLen;
				if ((x[j] < 0) || (x[j] > fAxisLen))
				{
					shall_gen = true;
					break;
				}
			}
		}

		for (int k = 0; k < nDimen; ++k)
		{
			fprintf(fp, " %.2f", x[k]);
			//fprintf(fp, " %f %f", x[k], x[k]);
		}
		fprintf(fp, "\n");
	}

	delete []x;
	fclose(fp);
}






/*****************************************************************
Multi-dimensional anti-correlated data
para
fname: output filename
n: cardinality (also decides the window size)
d: the dimensionality
u: universe
coded by yufei tao, jan 2005
modified by Xiaokui Xiao, 20061128
*****************************************************************/

void anti_correlated(const char *_fname, int _n, int _d, float _u)
{
	MTRand temp_rand;
	int k;

	temp_rand.seed();

	FILE *fp=fopen(_fname, "w");
	if (!fp)
	{
		printf("couldn't create file\n"); exit(1);
	}

	float *x=new float[_d];
	for (int j=0; j<_n; j++)
	{
		fprintf(fp, "%d", j+1);

		float tmp = -1;
		while (tmp<0 || tmp>1)
		{
			tmp = temp_rand.randNorm( 0.5, 0.05 * 0.05 );
		}
//			tmp = gaussian(0.5, 0.005);

		do
		{
			for (k=0; k<_d-1; k++)
			{
//				x[k] = new_uniform(0, 1);
				x[k] = temp_rand.rand();
			}
			x[_d-1] = tmp*_d;
			for (k=0; k<_d-1; k++)
				x[_d-1] -= x[k];
		} while (x[_d-1]<0 || x[_d-1]>1);

		for (k=0; k<_d; k++)
		{
			x[k] *= _u;
		}

		for (k=0; k<_d; k++)
		{
			fprintf(fp, " %.2f", x[k]);
			//fprintf(fp, " %f %f", x[k], x[k]);
		}
		fprintf(fp, "\n");
	}

	delete []x;
	fclose(fp);
}



/*****************************************************************
Multi-dimensional clustered dataset

fname: output filename
n: cardinality
d: the dimensionality
u: universe

cnum:	the number of clusters
csize:	the radius variance of each cluster
ccenter:	the centers of clusters

coded by Xiaokui Xiao, may 2005.
*****************************************************************/
void clustered_ds(char *_fname, int _n, int _d, float _u, int _cnum, float _csize, float** _ccenter )
{
	FILE *fp=fopen(_fname, "w");
	if (!fp)
	{
		printf("couldn't create file\n"); exit(1);
	}

	int pt_per_cluster = _n / _cnum;

	float* temp_cor = new float[_d];

	float temp_sum;

	bool shall_continue;

	float radius;

	int i, j, k; //, l; // commented by Raymond, 15 March 2007

	for (k = 0; k < _cnum; ++k)
	{
		for (i = 0; i < pt_per_cluster; i++)
		{
			radius = gaussian( 0, _csize );

			shall_continue = true;
			while (shall_continue == true)
			{
				temp_sum = 0;
				for (j = 0; j < _d; ++j)
				{
					temp_cor[j] = float(rand() - RAND_MAX / 2) / float(RAND_MAX) ;
					temp_sum += temp_cor[j] * temp_cor[j];
				}
				temp_sum = sqrt( temp_sum );

				shall_continue = false;
				for (j = 0; j < _d; ++j)
				{
					temp_cor[j] = _ccenter[k][j] + radius * temp_cor[j] / temp_sum;
//					temp_cor[j] = _ccenter[k][j] + float(rand() - RAND_MAX / 2) * 2.0 / float(RAND_MAX) * radius;
					if ((temp_cor[j] < 0) || (temp_cor[j] > _u))
					{
						shall_continue = true;
						break;
					}
				}
			}

//			std::cout << i << std::endl;

			fprintf(fp, "%d", i + 1 + k * pt_per_cluster);
			for (j = 0; j < _d; j++)
			{
				fprintf( fp, " %.2f", temp_cor[j] );
			}
			fprintf(fp, "\n");
		}
	}

	delete []temp_cor;

	fclose(fp);
}

// OData: Zipf distribution
void SPM_generateOData(char *filename, int noOfDim, int noOfTuple, float minValue, float maxValue, float zipfParameter)
{
	FILE *filePtr;
	int i, j;

	float aValue;

	filePtr = (FILE *) fopen(filename, "w");
	if (filePtr == NULL)
	{
		printf("File %s cannot be written!\n", filename);
		exit(0);
	}

	for (i = 0; i < noOfTuple; i++)
	{
		for (j = 0; j < noOfDim; j++)
		{
			aValue = minValue-1;
			while ((aValue < minValue) || (aValue > maxValue))
			{
				aValue = zipf(minValue, maxValue, (double) zipfParameter);
			}
#ifdef R_FLOAT
			fprintf(filePtr, "%.10f ", aValue);
#else
			fprintf(filePtr, "%d ", (int) aValue);	// make aValue to integer
#endif
		}
		fprintf(filePtr, "\n");
	}

	fclose(filePtr);
}

// PData: Gaussian distribution
void SPM_generatePData(char *filename, int noOfDim, int noOfTuple, float minValue, float maxValue, float sigma)
{
	FILE *filePtr;
	int i, j;

	float aValue;
	float mean;

	filePtr = (FILE *) fopen(filename, "w");
	if (filePtr == NULL)
	{
		printf("File %s cannot be written!\n", filename);
		exit(0);
	}

	mean = (maxValue-minValue)/2.0 + minValue;

	for (i = 0; i < noOfTuple; i++)
	{
		for (j = 0; j < noOfDim; j++)
		{
			aValue = minValue-1;
			while ((aValue < minValue) || (aValue > maxValue))
			{
				aValue = gaussian(mean, sigma);
			}
#ifdef R_FLOAT
			fprintf(filePtr, "%.10f ", aValue);
#else
			fprintf(filePtr, "%d ", (int) aValue);	// make aValue to integer
#endif
		}
		fprintf(filePtr, "\n");
	}

	fclose(filePtr);
}



R_TYPE **readData(char *filename, int noOfTuple, int noOfDim)
{
	R_TYPE **returnValue;
	int i, j;

	FILE *filePtr;

	filePtr = (FILE *) fopen(filename, "r");
	if (filePtr == NULL)
	{
		printf("File %s cannot be read!\n", filename);
		exit(0);
	}

	returnValue = (R_TYPE **) malloc(sizeof(R_TYPE *)*noOfTuple);
	for (i = 0; i < noOfTuple; i++)
	{
		returnValue[i] = (R_TYPE *) malloc(sizeof(R_TYPE)*noOfDim);
		memset(returnValue[i], 0, sizeof(R_TYPE)*noOfDim);

		for (j = 0; j < noOfDim; j++)
		{
#ifdef R_FLOAT
			fscanf(filePtr, "%f ", &(returnValue[i][j]));
#else
			fscanf(filePtr, "%d ", &(returnValue[i][j]));
#endif
		}

		fscanf(filePtr, "\n");
	}

	fclose(filePtr);

	return returnValue;
}

void SPM_generateWeightData_internal(int noOfDim, char *outFilename, int noOfTuple, R_TYPE **data, node_type *root, float radius, float sigma, rtree_info *rtreeInfo)
{
	FILE *filePtr;

	R_TYPE *query;

	int i, j;

	int noOfPointFound;
	int weight;

	float mean;

	filePtr = (FILE *) fopen(outFilename, "w");
	if (filePtr == NULL)
	{
		printf("File %s cannot be written!\n", outFilename);
		exit(0);
	}
	for (i = 0; i < noOfTuple; i++)
	{
		//printf("%d ..\n", i);
		query = data[i];

		noOfPointFound = 0;
		rectangle_search(root, query, radius, rtreeInfo, &noOfPointFound);

		mean = (float) noOfPointFound;

		weight = 0;
		while (weight <= 0)
		{
			weight = (int) gaussian(mean, sigma);
		}

		fprintf(filePtr, "%d ", weight);
		for (j = 0; j < noOfDim ; j++)
		{
#ifdef R_FLOAT
			fprintf(filePtr, "%.10f ", data[i][j]);
#else
			fprintf(filePtr, "%d ", data[i][j]);
#endif
		}
		fprintf(filePtr, "\n");
	}
	fclose(filePtr);
}

void build_tree_notFreeMemory(node_type **root, R_TYPE **data, int no_data, rtree_info *aInfo)
{ 
	int i, j;
	
	
	/* make tree root node first */
	
	tree_node_allocate(root, aInfo);
	
	for(i=0; i < aInfo->dim; i++) { 
		(*root)->b[i] = (R_TYPE)(-1 * INT_MAX);
       	(*root)->a[i] = (R_TYPE)(INT_MAX);
	}
	(*root)->id = -3;		//NO_ID
	(*root)->attribute=ROOT;
	(*root)->vacancy=aInfo->M;   
	(*root)->parent=NULL; 
	for(j=0; j < aInfo->M; j++)
		(*root)->ptr[j]=NULL;
	
	
	/* add data to the tree */
	for(i=0; i<no_data; i++) { 
		
#ifdef DEBUG
		if (i+1%1000==0)
			printf("insert data %d\n", i+1);
#endif
		
		if (i == 2)
		{
			printf("");
		}
		
		insert_node(*root, data[i], i, aInfo);  // i is the seq id.
		
		
		//free(data[i]);
	}
	
	//free(data);
	
} /*build_tree */


void SPM_generateWeightData(int noOfDim, char *POrgFile, char *OOrgFile, char *POutFile, char *OOutFile, int noOfTupleP, int noOfTupleO, rtree_info *rtreeInfo, float radius, float sigma)
{
	R_TYPE **OData;
	R_TYPE **PData;

	node_type *rootO;
	node_type *rootP;

	int i;

	OData = readData(OOrgFile, noOfTupleO, noOfDim);
	PData = readData(POrgFile, noOfTupleP, noOfDim);

	printf("Building tree for O...\n");
	build_tree_notFreeMemory(&rootO, OData, noOfTupleO, rtreeInfo);
	printf("Building tree for P...\n");
	build_tree_notFreeMemory(&rootP, PData, noOfTupleP, rtreeInfo);

	// O
	SPM_generateWeightData_internal(noOfDim, OOutFile, noOfTupleO, OData, rootO, radius, sigma, rtreeInfo);

	// P
	SPM_generateWeightData_internal(noOfDim, POutFile, noOfTupleP, PData, rootP, radius, sigma, rtreeInfo);

	// free the array
	for (i = 0; i < noOfTupleO; i++)
	{
		free(OData[i]);
	}
	free(OData);
	for (i = 0; i < noOfTupleP; i++)
	{
		free(PData[i]);
	}
	free(PData);
	free(rtreeInfo);

}

void SPM_generateWeightData_zeroSigma_internal(char *filename, int noOfTuple, int noOfDim, R_TYPE **data)
{
	FILE *filePtr;
	int i, j;

	filePtr = (FILE *) fopen(filename, "w");
	if (filePtr == NULL)
	{
		printf("File %s cannot be written!\n");
		exit(0);
	}

	for (i = 0; i < noOfTuple; i++)
	{
		fprintf(filePtr, "1 ");
		for (j = 0; j < noOfDim; j++)
		{
#ifdef R_FLOAT
			fprintf(filePtr, "%.10f ", data[i][j]);
#else
			fprintf(filePtr, "%d ", data[i][j]);
#endif
		}
		fprintf(filePtr, "\n");
	}
	fclose(filePtr);
}

void SPM_generateWeightData_zeroSigma(int noOfDim, char *POrgFile, char *OOrgFile, char *POutFile, char *OOutFile, int noOfTupleP, int noOfTupleO, rtree_info *rtreeInfo, float radius, float sigma)
{
	R_TYPE **OData;
	R_TYPE **PData;

	int i;

	OData = readData(OOrgFile, noOfTupleO, noOfDim);
	PData = readData(POrgFile, noOfTupleP, noOfDim);

	
	// O
	SPM_generateWeightData_zeroSigma_internal(OOutFile, noOfTupleO, noOfDim, OData);

	// P
	SPM_generateWeightData_zeroSigma_internal(POutFile, noOfTupleP, noOfDim, PData);

	// free the array
	for (i = 0; i < noOfTupleO; i++)
	{
		free(OData[i]);
	}
	free(OData);
	for (i = 0; i < noOfTupleP; i++)
	{
		free(PData[i]);
	}
	free(PData);
	free(rtreeInfo);

}


void printUsage(char *command)
{
	printf(" Usage:\n");
	printf("   %s genMode <... remaining parameters depends on genMode ...>\n", command);
	printf("     genMode = 1: generate synthetic points\n");
	printf("     genMode = 2: generate synthetic weights from existing data points\n");
	printf("   e.g.\n");
	printf("   %s 1 noOfDim noOfTupleP noOfTupleO Pfile Ofile minValue maxValue zipfParameter sigma\n", command);
	printf("   %s 2 noOfDim noOfTupleP noOfTupleO POrgFile OOrgFile POutFile OOutFile radius sigma\n", command);
}

// 1 3 10 8 pData.txt oData.txt 0 10000 0.8 5000
// 2 3 10 8 pData.txt oData.txt pWeightData.txt oWeightData.txt 500 10
int main(int argc, char **argv)
{
	int noOfDim;
	int noOfTupleP;
	int noOfTupleO;

	char *Pfile;
	char *Ofile;

	char *POrgFile;
	char *OOrgFile;
	char *POutFile;
	char *OOutFile;
	
	float minValue;
	float maxValue;
	float zipfParameter;
	float sigma;

	float radius;

	int genMode;

	rtree_info *rtreeInfo;



	if (argc < 2)
	{
		printUsage(argv[0]);
		exit(0);
	}

	genMode = atoi(argv[1]);

	if (genMode == 1)
	{
		// genMode = 1
		//    generate synthetic points

		if (argc != 11)
		{
			printUsage(argv[0]);
			exit(0);
		}

		// parameters
		noOfDim = atoi(argv[2]);
		noOfTupleP = atoi(argv[3]);
		noOfTupleO = atoi(argv[4]);
		Pfile = argv[5];
		Ofile = argv[6];
		minValue = atof(argv[7]);
		maxValue = atof(argv[8]);
		zipfParameter = atof(argv[9]);
		sigma = atof(argv[10]);

		// data generation
		printf("Generating O data points...\n");
		SPM_generateOData(Ofile, noOfDim, noOfTupleO, minValue, maxValue, zipfParameter);
		printf("Generating P data points...\n");
		SPM_generatePData(Pfile, noOfDim, noOfTupleP, minValue, maxValue, sigma);
	}
	else if (genMode == 2)
	{
		// genMode = 2
		//     generate synthetic weights from existing data points

		if (argc != 11)
		{
			printUsage(argv[0]);
			exit(0);
		}

		// parameters
		noOfDim = atoi(argv[2]);
		noOfTupleP = atoi(argv[3]);
		noOfTupleO = atoi(argv[4]);
		POrgFile = argv[5];
		OOrgFile = argv[6];
		POutFile = argv[7];
		OOutFile = argv[8];
		radius = atof(argv[9]);
		sigma = atof(argv[10]);

		rtreeInfo = rtreeInfo_new();

		rtreeInfo->dim = noOfDim;
		rtreeInfo->M = 130;
		rtreeInfo->no_histogram = noOfTupleP;	// dummy

		rtreeInfo->m = rtreeInfo->M/2;
		rtreeInfo->extra_level = 0;
		rtreeInfo->reinsert_p = rtreeInfo->m;

		// weight generation
		printf("Generating weights...\n");
		if (sigma == 0)
		{
			SPM_generateWeightData_zeroSigma(noOfDim, POrgFile, OOrgFile, POutFile, OOutFile, noOfTupleP, noOfTupleO, rtreeInfo, radius, sigma);
		}
		else
		{
			SPM_generateWeightData(noOfDim, POrgFile, OOrgFile, POutFile, OOutFile, noOfTupleP, noOfTupleO, rtreeInfo, radius, sigma);
		}

	}

	return 0;
}
